// This is the main DLL file.


#include <Windows.h>

#include "thundervideo.h"

#include <gcroot.h>

extern "C" {
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libswscale/swscale.h>
#include <libavutil/imgutils.h>
}

using namespace thundervideo;

namespace {
    const char sdpFormat[] = "v=0\n"
        "m=video %d RTP/AVP 96\n"
        "c=IN IP4 0.0.0.0\n"
        "a=rtpmap:96 H264/90000\n";


}


namespace thundervideo {
    class ThunderVideoNativeCtx {
    private:
        AVFormatContext *fmtCtx;
        AVCodecContext *codecCtx;
        SwsContext *swsCtx;
        AVIOContext *ioCtx;
        AVPacket *pkt;
        AVFrame *frame;
        AVFrame *outFrame;
        bool pktFullyProcessed;
        int w, h;
        AVPixelFormat pixfmt;
        uint8_t *outbuf;
        int readOffset;
        // 5 for 5-character port
        char sdp[sizeof(sdpFormat) + 5];

        long timeoutMillis;
        gcroot<System::Diagnostics::Stopwatch ^> timer;


        static int callback(void *opaque) {
            ThunderVideoNativeCtx *self = (ThunderVideoNativeCtx *)opaque;
            return self->interrupt_callback();
        }

        int interrupt_callback()
        {
            if (timer->ElapsedMilliseconds >= timeoutMillis)
                return 1;
            return 0;
        }

        static int read_pkt(void *opaque, uint8_t *buf, int buf_size)
        {
            ThunderVideoNativeCtx *self = (ThunderVideoNativeCtx *)opaque;
            return self->read(buf, buf_size);
        }

        int read(uint8_t *buf, int buf_size)
        {
            int r = strlen(sdp) - readOffset;
            if (!r)
                return AVERROR_EOF;
            if (r > buf_size)
                r = buf_size;
            memcpy(buf, sdp + readOffset, r);
            readOffset += r;
            return r;
        }


    public:
        ThunderVideoNativeCtx() : fmtCtx(NULL), codecCtx(NULL), swsCtx(NULL),
                                  ioCtx(NULL), 
                                  pkt(NULL), frame(NULL), pktFullyProcessed(true),
                                  w(-1), h(-1), pixfmt(AV_PIX_FMT_NONE),
                                  outbuf(NULL), readOffset(0)
        {
            pkt = av_packet_alloc();
            frame = av_frame_alloc();
            outFrame = av_frame_alloc();
            timer = gcnew System::Diagnostics::Stopwatch();
        }

        ~ThunderVideoNativeCtx()
        {
            close();
            av_packet_free(&pkt);
            av_frame_free(&frame);
            av_frame_free(&outFrame);
        }


        ThunderVideoError open(int port, int timeoutMs)
        {
            try {
                if (fmtCtx || port < 0 || port > 65535)
                    throw 1;

                AVDictionary *d = NULL;
                av_dict_set(&d, "protocol_whitelist", "file,rtp,udp", 0);

                this->timeoutMillis = timeoutMs;
                timer->Restart();

#if 0
                // XXX - use port, dealloc on error
                if (avformat_open_input(&fmtCtx, "test.sdp", NULL, &d) != 0) {
                    fflush(stdout);
                    fflush(stderr);
                    throw 2;
                }
                fmtCtx->interrupt_callback.callback = callback;
                fmtCtx->interrupt_callback.opaque = this;
                av_dict_free(&d);
#endif
                // Write out SDP with given port
                sprintf(sdp, sdpFormat, port);

                // ioctx cleanup
                // ioctx buffer cleanup
                // fmtCtx proper cleanup on both error and normal close
                readOffset = 0;
                uint8_t *buffer = (uint8_t *)av_malloc(4096);
                ioCtx = avio_alloc_context(buffer, 4096, 0, this, read_pkt, NULL, NULL);
                fmtCtx = avformat_alloc_context();
                fmtCtx->pb = ioCtx;
                fmtCtx->interrupt_callback.callback = callback;
                fmtCtx->interrupt_callback.opaque = this;
                fmtCtx->flags |= AVFMT_FLAG_CUSTOM_IO;
                AVInputFormat *infmt = av_find_input_format("sdp");

                int rc = avformat_open_input(&fmtCtx, "", infmt, &d);
                av_dict_free(&d);

                if (rc != 0) {
                    av_freep(&ioCtx->buffer);
                    avio_context_free(&ioCtx);
                    avformat_free_context(fmtCtx);
                    fmtCtx = NULL;
                    throw 2;
                }

                rc = avformat_find_stream_info(fmtCtx, NULL);
                if (rc < 0)
                    throw 3;

                if (fmtCtx->nb_streams != 1)
                    throw 4;

                AVCodec *codec = avcodec_find_decoder(fmtCtx->streams[0]->codecpar->codec_id);
                if (codec == NULL)
                    return ThunderVideoError::FAILED_OTHER;
                codecCtx = avcodec_alloc_context3(codec);
                if (!codecCtx)
                    throw 5;

                if (avcodec_parameters_to_context(codecCtx, fmtCtx->streams[0]->codecpar) < 0)
                    throw 6;
                codecCtx->time_base = codecCtx->pkt_timebase =
                    fmtCtx->streams[0]->time_base;

                if (avcodec_open2(codecCtx, codec, NULL) < 0)
                    throw 7;

                w = h = -1;
                pixfmt = AV_PIX_FMT_NONE;
            } catch (int &x) {
                close();
                printf("Open Fail %d\n", x);
                if (timer->ElapsedMilliseconds > timeoutMillis)
                    return ThunderVideoError::FAILED_TIMEOUT;
                else
                    return ThunderVideoError::FAILED_OTHER;
            }
            

            return ThunderVideoError::OK;
        }

        void close()
        {
            if (codecCtx)
                avcodec_free_context(&codecCtx);
            if (fmtCtx)
                avformat_close_input(&fmtCtx);
            if (swsCtx) {
                sws_freeContext(swsCtx);
                swsCtx = NULL;
            }

            if (ioCtx) {
                av_freep(&ioCtx->buffer);
                avio_context_free(&ioCtx);
            }

            if (!pktFullyProcessed)
                av_packet_unref(pkt);
            pktFullyProcessed = true;
            av_frame_unref(frame);

            av_frame_unref(outFrame);
            delete[] outbuf;
            outbuf = NULL;
        }

        ThunderVideoError getNextFrame(Bitmap ^%image, int timeoutMs)
        {
            if (!fmtCtx)
                return ThunderVideoError::FAILED_OTHER;

            bool gotFrame = false;

            this->timeoutMillis = timeoutMs;
            timer->Restart();

            while (!gotFrame) {
                try {
                    if (pktFullyProcessed) {
                        int rc = av_read_frame(fmtCtx, pkt);
                        if (rc == AVERROR_EXIT)
                            throw 10;
                        else if (rc < 0)
                            throw 1;

                        rc = avcodec_send_packet(codecCtx, pkt);
                        av_packet_unref(pkt);
                        switch (rc) {
                        case 0:
                            // success
                            break;
                        case AVERROR_EXIT:
                            throw 10;
                        case AVERROR(EAGAIN):
                        case AVERROR_EOF:
                        default:
                            throw 1;
                        }

                        pktFullyProcessed = false;
                    }

                    switch (avcodec_receive_frame(codecCtx, frame))
                    {
                    case 0:
                        gotFrame = true;
                        break;
                    case AVERROR(EAGAIN):
                        pktFullyProcessed = true;
                        break;
                    case AVERROR_EXIT:
                        throw 10;
                    case AVERROR_EOF:
                    case AVERROR(EINVAL):
                        throw 1;
                    default:
                        // decode error - move on
                        pktFullyProcessed = true;
                        break;
                    }
                }
                catch (int &x) {
                    if (x != 10)
                        return ThunderVideoError::FAILED_OTHER;
                }
                if (timer->ElapsedMilliseconds > timeoutMs)
                    return ThunderVideoError::FAILED_TIMEOUT;

            }

            // We have a frame
            if (w != frame->width || h != frame->height || pixfmt != frame->format) {
                if (swsCtx)
                    sws_freeContext(swsCtx);
                swsCtx = sws_getContext(frame->width, frame->height, 
                    (AVPixelFormat)frame->format,
                    frame->width, frame->height, AV_PIX_FMT_BGRA, 0, NULL, NULL, NULL);
                if (!swsCtx)
                    return ThunderVideoError::FAILED_OTHER;
                w = frame->width;
                h = frame->height;
                pixfmt = (AVPixelFormat)frame->format;

                int outbufSize = av_image_get_buffer_size(AV_PIX_FMT_BGRA, w, h, 1);
                delete[] outbuf;
                outbuf = new uint8_t[outbufSize];
                av_image_fill_arrays(outFrame->data, outFrame->linesize, outbuf, AV_PIX_FMT_BGRA, w, h, 1);
            }
            sws_scale(swsCtx, frame->data, frame->linesize, 0, h, outFrame->data, outFrame->linesize);

            Bitmap ^bm = gcnew Bitmap(frame->width, frame->height, outFrame->linesize[0], Imaging::PixelFormat::Format32bppArgb, IntPtr(outFrame->data[0]));

            image = bm;

            return ThunderVideoError::OK;
        }


    };
}


ThunderVideo::ThunderVideo()
{
    static bool hasBeenInit = false;
    if (!hasBeenInit) {
        hasBeenInit = staticInit();
    }

    nativeCtx = new ThunderVideoNativeCtx();
}

ThunderVideo::~ThunderVideo()
{
    this->!ThunderVideo();
}

ThunderVideo::!ThunderVideo()
{
    delete nativeCtx;
    nativeCtx = NULL;
}

// Return values:
// OK: All good, ready to go
// FAILED_PORT_IN_USE:  Specified port could not be opened, already in use
// FAILED_OTHER: Already open or some other error occurred
ThunderVideoError ThunderVideo::open(int port, int timeoutMs)
{
    return nativeCtx->open(port, timeoutMs);
}

// Safe to call in any state; no-op if not open
void ThunderVideo::close()
{
    nativeCtx->close();
}

// Try to read and decode the next frame, waiting up to the given # of milliseconds
// Returns Bitmap of next frame
// Return values:
// OK: All good, image data is set
// FAILED_TIMEOUT: timeoutMs went by without any new data or new frame being decoded
// FAILED_OTHER: Any other error
ThunderVideoError ThunderVideo::getNextFrame(Bitmap ^%image, int timeoutMs)
{
    return nativeCtx->getNextFrame(image, timeoutMs);
}


bool ThunderVideo::staticInit()
{
 //   av_register_all();
    avformat_network_init();
    return true;
}